import hashlib
import logging
import time
from typing import Any, Optional

from typeguard import typechecked

from psi_kinesis.kinesis_reader import KinesisReader, PROVISIONED_THROUGHPUT_EXCEPTION

MAX_FAILED_RECORDS = 25  # max number of failed put records to raise
MAX_INVALID_RECORDS = 10  # max number of invalid put records to raise
MAX_PARTITION_KEY_LEN = 256  # max number of characters in a partition key
MAX_PUT_RECORDS_BATCH = 500  # max number of records in put records operation


@typechecked
class KinesisWriter(KinesisReader):
    """KinesisWriter provides read/write utility functions for interacting with Kinesis resources"""

    def __init__(
        self,
        aws_session: Optional[Any] = None,
        max_retry_count: Optional[int] = None,
        retry_backoff_rate: Optional[float] = None,
        logger: Optional[logging.Logger] = None,
    ):
        """Initializes a KinesisWriter

        Parameters
        ----------
        aws_session : Any
            (optional) A boto3.session.Session() object used to create boto3 clients.

        max_retry_count : int
            (optional) The max number of retries for Kinesis operations like
            put_records or get_records. Default value of 5.

        retry_backoff_rate : float
            (optional) The retry backoff rate multiplier for Kinesis operations like
            put_records or get_records. Default value of 3.0.

        logger : logging.Logger
            (optional) The logger used by a KinesisWriter object. If unspecified,
            a new logger will be created.
        """
        super().__init__(aws_session, max_retry_count, retry_backoff_rate, logger)
        # placeholder for __init__ differences between a KinesisReader and a KinesisWriter

    def put_records(
        self, stream_name: str, records: Optional[list] = None, quit_on_error: bool = False
    ) -> list[dict]:
        """Puts records into a Kinesis stream.

        This function handles the following:
          1. Validating each record's format
          2. Backoff and retry in the event of a ProvisionedThroughputExceededException
          3. Batching (the PutRecords API has a maximum of 500 records per operation)
          4. Partitioning (see "Partitioning" below)
          5. Gracefully handles failed records and exceptions

        ### Partitioning

        Kinesis streams divide data among shards using either a partition key or an
        explicit hash key to determine which shard a record goes to.
        See: https://docs.aws.amazon.com/kinesis/latest/APIReference/API_PutRecords.html

        Either a "PartitionKey" or an "ExplicitHashKey" may be placed on each record
        dictionary to control which shard the record goes to.
        If neither the "PartitionKey" nor "ExplicitHashKey" are set on the record, this
        function will attempt to automatically distribute the records uniformly among shards.

        Parameters
        ----------
        stream_name : str
            The name of the Kinesis stream where the records will be sent

        records : list
            list of records, where each record is a dictionary like:

            {
                'PartitionKey': 'string',
                'ExplicitHashKey': 'string',
                'Data': b'bytes'
            }

            The "Data" field is required and must be bytes.
            The "PartitionKey" and "ExplicitHashKey" are optional. If one is set,
            do not include the other.

        quit_on_error : bool
            (optional) Whether to quit putting records if an error occurs or to
            attempt writing all records (even if an error occurs). Default False.

        Returns
        -------
        list[dict]
            A list of locations of the new records with the form:

            [
                {
                    'ShardId': 'string',
                    'SequenceNumber': 'string'
                }
            ]
        """
        # Validate records
        invalid_records = self.__validate_put_records(records)
        assert not invalid_records, f"Some records are invalid: {invalid_records}"
        self.logger.info(f'Putting {len(records)} record(s) into Kinesis stream "{stream_name}"')

        # Batch records
        batches = self.__batch_put_records(records)

        # Write each batch
        results, exceptions = [], []
        for batch in batches:
            try:
                batch = self.__partition_put_records(batch)
                resp = self.__put_records_batch(batch, stream_name)
                results.extend(resp.get("Records") or [])
                if quit_on_error and resp.get("FailedRecordCount"):
                    break
            except Exception as e:
                self.logger.error(f"Failed to put Kinesis record with exception: {e}")
                if quit_on_error:
                    raise e
                exceptions.append(e)

        # Raise any exceptions that occurred
        if exceptions:
            self.logger.error("At least one exception occurred while putting records")
            raise Exception(exceptions)

        # Assert no records failed to write
        failed_records = [r for r in results if r.get("ErrorMessage")]
        assert (
            not failed_records
        ), f"{len(failed_records)} record(s) failed to write: {failed_records[:MAX_FAILED_RECORDS]}"

        # Return the successfully written record locations
        return [
            {"SequenceNumber": r.get("SequenceNumber"), "ShardId": r.get("ShardId")}
            for r in results
        ]

    def __put_records_batch(
        self,
        batch: list[dict],
        stream_name: str,
        wait_seconds: Optional[float] = None,
        retry_count: Optional[int] = None,
    ) -> dict[str, Any]:
        """Performs the actual boto3 put_records. Recursively handles backoff
        and retry in the event of a ProvisionedThroughputExceededException(s)."""
        self.logger.debug(f"Putting batch with {len(batch)} record(s)")
        try:
            assert int(retry_count or 0) <= self.max_retry_count, "Retries have been exhausted"
            if wait_seconds:
                self.logger.info(f"Waiting {wait_seconds} second(s)...")
                time.sleep(round(wait_seconds, 2))

            resp = self.client.put_records(Records=batch, StreamName=stream_name)

            has_throughput_exception, has_other_exception = (
                self.__put_records_resp_has_throughput_exception(resp=resp)
            )
            if has_other_exception:
                self.logger.error("Encountered an unexpected Exception. Skipping retry.")
                return resp
            if has_throughput_exception:
                self.logger.warning(
                    "Encountered a ProvisionedThroughputExceededException. Retrying."
                )
                return self.__put_records_batch(
                    batch=batch,
                    stream_name=stream_name,
                    wait_seconds=self._next_wait_seconds(wait_seconds),
                    retry_count=self._next_retry_count(retry_count),
                )
            return resp
        except Exception as e:
            self.logger.error(f"Failed to put Kinesis records batch with exception: {e}")
            raise e

    def __partition_put_records(self, records: list[dict]) -> list[dict]:
        """Assigns a new partition key to records without an existing
        partition key or explicit hash key"""
        for r in records:
            if not ("PartitionKey" in r or "ExplicitHashKey" in r):
                sha256 = hashlib.sha3_256()
                sha256.update(r["Data"])
                r["PartitionKey"] = sha256.hexdigest()
        return records

    def __batch_put_records(
        self, records: list[dict], batch_size: int = MAX_PUT_RECORDS_BATCH
    ) -> Any:
        """Chunks the list of records into chunks of size batch_size"""
        for i in range(0, len(records), batch_size):
            yield records[i : i + batch_size]

    def __validate_put_records(self, records: list) -> list[dict]:
        """Validates records to write to a Kinesis stream.
        Returns a list of dictionaries of invalid records with the form:

        [
            {
                'record': Any,     # The invalid record
                'reason': 'string' # Why the record is invalid
            }
        ]

        The number of invalid records returned is limited by MAX_INVALID_RECORDS.
        """
        invalid_records = []
        for r in records:
            is_valid, invalid_reason = self.__validate_put_record(r)
            if not is_valid:
                invalid_records.append({"record": r, "invalid_reason": invalid_reason})
            if len(invalid_records) >= MAX_INVALID_RECORDS:
                break
        return invalid_records

    def __validate_put_record(self, record) -> tuple[bool, str]:
        """Validates a record to write to a Kinesis stream.
        Returns a tuple[bool, str] with two elements:
          1. bool : Whether the record is valid
          2. str : Reason for why the record is invalid. Empty if the record is valid
        """
        try:
            assert isinstance(record, dict), "Expected record to be a dictionary"
            assert isinstance(record.get("Data"), bytes), "Expected record.Data to be bytes"

            has_partiton_key = "PartitionKey" in record
            has_explicit_hash_key = "ExplicitHashKey" in record
            assert not (
                has_partiton_key and has_explicit_hash_key
            ), "Expected either record.PartitionKey or record.ExplicitHashKey, not both"
            if has_partiton_key:
                assert isinstance(
                    record["PartitionKey"], str
                ), "Expected record.PartitionKey to be a string"
                assert (
                    len(record["PartitionKey"]) <= MAX_PARTITION_KEY_LEN
                ), f"Expected record.PartitionKey to have fewer than {MAX_PARTITION_KEY_LEN + 1} characters"
            if has_explicit_hash_key:
                assert isinstance(
                    record["ExplicitHashKey"], str
                ), "Expected record.ExplicitHashKey to be a string"
            return True, ""
        except AssertionError as e:
            return False, str(e)

    def __put_records_resp_has_throughput_exception(self, resp: Any) -> tuple[bool, bool]:
        """Checks for ProvisionedThroughputExceededException(s) in a put_records response"""
        has_throughput_exception, has_other_exception = False, False
        if isinstance(resp, dict) and resp.get("FailedRecordCount"):
            for r in resp.get("Records") or []:
                if r.get("ErrorCode") == PROVISIONED_THROUGHPUT_EXCEPTION:
                    has_throughput_exception = True
                elif r.get("ErrorCode"):
                    has_other_exception = True
        return has_throughput_exception, has_other_exception



import logging
import time
from datetime import datetime
from typing import Any, Optional

import boto3
import botocore
from botocore.exceptions import ClientError
from typeguard import typechecked

from psi_kinesis import get_logger

DEFAULT_MAX_RETRY_COUNT = 5  # default max number of retries for Kinesis operations
DEFAULT_RETRY_BACKOFF_RATE = 3.0  # default retry backoff multiplier
PROVISIONED_THROUGHPUT_EXCEPTION = "ProvisionedThroughputExceededException"
VALID_SHARD_ITERATOR_TYPES = [
    "AT_SEQUENCE_NUMBER",
    "AFTER_SEQUENCE_NUMBER",
    "TRIM_HORIZON",
    "LATEST",
    "AT_TIMESTAMP",
]


@typechecked
class KinesisReader:
    """KinesisReader provides read-only utility functions for interacting with Kinesis resources"""

    def __init__(
        self,
        aws_session: Optional[Any] = None,
        max_retry_count: Optional[int] = None,
        retry_backoff_rate: Optional[float] = None,
        logger: Optional[logging.Logger] = None,
    ):
        """Initializes a KinesisReader

        Parameters
        ----------
        aws_session : Any
            (optional) A boto3.session.Session() object used to create boto3 clients.

        max_retry_count : int
            (optional) The max number of retries for Kinesis operations like
            get_records. Default value of 5.

        retry_backoff_rate : float
            (optional) The retry backoff rate multiplier for Kinesis operations like
            get_records. Default value of 3.0.

        logger : logging.Logger
            (optional) The logger used by a KinesisReader object. If unspecified,
            a new logger will be created.
        """
        assert (
            max_retry_count is None or max_retry_count >= 0
        ), f"Max retry count must be greater than or equal to 0"
        assert (
            retry_backoff_rate is None or retry_backoff_rate > 0
        ), f"Retry backoff rate must be greater than 0"
        self.__aws_session = aws_session or boto3.session.Session()
        self.__original_aws_session = self.aws_session
        self.__client = self.aws_session.client("kinesis")
        self.__max_retry_count = (
            DEFAULT_MAX_RETRY_COUNT if max_retry_count is None else max_retry_count
        )
        self.__retry_backoff_rate = retry_backoff_rate or DEFAULT_RETRY_BACKOFF_RATE
        self.__logger = logger or get_logger()

    @property
    def aws_session(self) -> Any:
        return self.__aws_session

    @property
    def client(self) -> Any:
        return self.__client

    @property
    def max_retry_count(self) -> int:
        return self.__max_retry_count

    @property
    def retry_backoff_rate(self) -> float:
        return self.__retry_backoff_rate

    @property
    def logger(self) -> logging.Logger:
        return self.__logger

    def assume_role(
        self,
        role_name: str,
        account_id: str,
        region: str = "us-east-1",
        role_session_name: str = "CrossAccountKinesisRole",
    ):
        """Assume an IAM role to perform subsequent Kinesis actions

        Parameters
        ----------
        role_name : str
            Name of the IAM role to assume

        account_id : str
            AWS account ID where the IAM role resides

        region : str
            (optional) AWS region code to use as the default
            region for subsequent Kinesis actions

        role_session_name : str
            (optional) An identifier for the assumed role session.
            Default value "CrossAccountKinesisRole"
        """
        self.logger.info(
            f"Assuming role arn:aws:iam::{account_id}:role/{role_name} with default region {region}"
        )
        try:
            sts_client = self.__original_aws_session.client("sts")
            role_creds = sts_client.assume_role(
                RoleArn=f"arn:aws:iam::{account_id}:role/{role_name}",
                RoleSessionName=role_session_name,
            )["Credentials"]

            self.__aws_session = boto3.Session(
                aws_access_key_id=role_creds["AccessKeyId"],
                aws_secret_access_key=role_creds["SecretAccessKey"],
                aws_session_token=role_creds["SessionToken"],
                region_name=region,
            )
            self.__client = self.aws_session.client("kinesis", region_name=region)
        except Exception as e:
            self.logger.error(f"Failed to assume role with exception: {e}")
            raise e

    def does_stream_exist(self, stream_name: str) -> bool:
        """Checks if a Kinesis stream exists and returns true/false

        Parameters
        ----------
        stream_name : str
            The name of the stream to look for

        Returns
        -------
        bool
            Whether or not the stream exists
        """
        try:
            self.logger.info(f'Describing stream with name "{stream_name}"')
            resp = self.client.describe_stream(StreamName=stream_name)
            return resp.get("StreamDescription", {}).get("StreamName") == stream_name
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "ResourceNotFoundException":
                return False
            self.logger.error(f"Failed to describe Kinesis stream with exception: {e}")
            raise e
        except Exception as e:
            self.logger.error(f"Failed to describe Kinesis stream with exception: {e}")
            raise e

    def list_shards(
        self, stream_name: str, shard_filter: Optional[dict[str, Any]] = None
    ) -> list[dict]:
        """list shards for a given Kinesis stream. Handles paginiation.

        Parameters
        ----------
        stream_name : str
            The name of the Kinesis stream to retrieve shards

        shard_filter : dict[str, Any]
            (optional) Allows filtering the response of the
            ListShards API with a dictionary like:

            {
                'Type': 'string',
                'ShardId': 'string',
                'Timestamp': datetime(2015, 1, 1)
            }

        Returns
        -------
        list[dict]
            The list of shards for the given Kinesis stream.
            Each shard is a dictionary like:

            {
                'ShardId': 'string',
                'ParentShardId': 'string',
                'AdjacentParentShardId': 'string',
                'HashKeyRange': {
                    'StartingHashKey': 'string',
                    'EndingHashKey': 'string'
                },
                'SequenceNumberRange': {
                    'StartingSequenceNumber': 'string',
                    'EndingSequenceNumber': 'string'
                }
            }
        """
        try:
            self.logger.info(f'Listing shards for Kinesis stream "{stream_name}"')
            list_shards_input = {"StreamName": stream_name}
            if shard_filter:
                list_shards_input["ShardFilter"] = shard_filter

            shards, has_more_shards = [], True
            while has_more_shards:
                resp = self.client.list_shards(**list_shards_input)
                shards.extend(resp.get("Shards") or [])
                list_shards_input["NextToken"] = resp.get("NextToken")
                has_more_shards = bool(resp.get("NextToken"))
            return shards
        except Exception as e:
            self.logger.error(f"Failed to list shards for Kinesis stream with exception: {e}")
            raise e

    def get_records(
        self,
        stream_name: str,
        shard_id: Optional[str] = None,
        shard_iterator_type: Optional[str] = None,
        starting_sequence_number: Optional[str] = None,
        timestamp: Optional[datetime] = None,
        shard_filter: Optional[dict[str, Any]] = None,
    ) -> dict[str, list]:
        """Retrieve records from a Kinesis stream. By default, retrieves all records
        in the stream.
        You can limit which shards to retrieve records from using
        either shard_id or shard_filter (but not both).
        You can control how records are iterated with shard_iterator_type,
        starting_sequence_number, and timestamp. For more information, see:
        https://docs.aws.amazon.com/kinesis/latest/APIReference/API_GetShardIterator.html

        This function handles Backoff and retry in the event of a
        ProvisionedThroughputExceededException.

        Parameters
        ----------
        stream_name : str
            The name of the Kinesis stream to get records from

        shard_id : str
            (optional) The ID of the shard to get records from.
            If set, shard_filter cannot be specified.

        shard_iterator_type : str
            (optional) Determines how the shard iterator is used to start reading records
            from the shard.
            Defaults to "TRIM_HORIZON"

        starting_sequence_number : str
            (optional) The sequence number of the record in the shard from which to start
            reading.
            If set, timestamp cannot be specified.

        timestamp : datetime
            (optional) The timestamp of the record from which to start reading.
            If set, starting_sequence_number cannot be specified.

        shard_filter : dict[str, Any]
            (optional) Allows filtering the response of the
            ListShards API with a dictionary like:

            {
                'Type': 'string',
                'ShardId': 'string',
                'Timestamp': datetime(2015, 1, 1)
            }

            If set, shard_id cannot be specified

        Returns
        -------
        dict[str, list]
            Dictionary mapping shard ID to list of records with the form:

            {
                'string': [
                    {
                        'SequenceNumber': 'string',
                        'ApproximateArrivalTimestamp': datetime(2015, 1, 1),
                        'Data': b'bytes',
                        'PartitionKey': 'string',
                        'EncryptionType': 'NONE'|'KMS'
                    }
                ]
            }
        """
        self.logger.info(f"Getting records from Kinesis stream: {stream_name}")

        # Validate input
        shard_iterator_type = shard_iterator_type or "TRIM_HORIZON"
        assert not (
            shard_id and shard_filter
        ), "Cannot get records with both a shard ID and a shard filter. Specify one or the other"
        assert not (
            starting_sequence_number and timestamp
        ), "Cannot specify both timestamp and starting sequence number. Specify one or the other"
        assert (
            shard_iterator_type in VALID_SHARD_ITERATOR_TYPES
        ), f'Invalid shard iterator type "{shard_iterator_type}". Must be one of: {VALID_SHARD_ITERATOR_TYPES}'
        if shard_iterator_type in ["AT_SEQUENCE_NUMBER", "AFTER_SEQUENCE_NUMBER"]:
            assert (
                starting_sequence_number
            ), "Starting sequence number must be specified if shard iterator type is one of [AT_SEQUENCE_NUMBER,AFTER_SEQUENCE_NUMBER]"
        elif shard_iterator_type in ["AT_TIMESTAMP"]:
            assert (
                timestamp
            ), "A timestamp must be specified if shard iterator type is one of [AT_TIMESTAMP]"
        else:
            assert not (
                starting_sequence_number or timestamp
            ), "Cannot specify a starting sequence number or timestamp if shard iteratory type is one of [TRIM_HORIZON,LATEST]"

        # Get shard IDs
        if shard_id:
            shard_ids = [shard_id]
        else:
            shard_ids = [
                s["ShardId"] for s in self.list_shards(stream_name, shard_filter=shard_filter)
            ]

        # For each shard, get a shard iterator and retrieve records using the iterator
        records = {}
        for sid in shard_ids:
            shard_iterator = self.__get_shard_iterator(
                stream_name=stream_name,
                shard_id=sid,
                shard_iterator_type=shard_iterator_type,
                starting_sequence_number=starting_sequence_number,
                timestamp=timestamp,
            )
            records[sid] = self.get_records_with_shard_iterator(shard_iterator) or []
        return records

    def get_records_with_shard_iterator(self, shard_iterator: Optional[str] = None) -> list[dict]:
        """Retrieve records from a Kinesis stream's shard beginning at shard_iterator.
        This function handles backoff and retry in the event of a
        ProvisionedThroughputExceededException.

        Parameters
        ----------
        shard_iterator : str
            The position in the shard from which to start sequentially reading records.

        Returns
        -------
        list[dict]
            list of records of the form

            [
                {
                    'SequenceNumber': 'string',
                    'ApproximateArrivalTimestamp': datetime(2015, 1, 1),
                    'Data': b'bytes',
                    'PartitionKey': 'string',
                    'EncryptionType': 'NONE'|'KMS'
                }
            ]
        """
        self.logger.info(f"Getting records with shard iterator: {shard_iterator}")
        records, has_more_records = [], True
        while has_more_records:
            self.logger.debug(f"Getting records with shard iterator: {shard_iterator}")
            resp = self.__get_records_with_shard_iterator(shard_iterator)
            records.extend(resp.get("Records") or [])
            shard_iterator = resp.get("NextShardIterator")
            has_more_records = resp.get("MillisBehindLatest") > 0 and not resp.get("ChildShards")
        return records

    def __get_shard_iterator(
        self,
        stream_name: str,
        shard_id: Optional[str] = None,
        shard_iterator_type: Optional[str] = None,
        starting_sequence_number: Optional[str] = None,
        timestamp: Optional[datetime] = None,
        wait_seconds: Optional[float] = None,
        retry_count: Optional[int] = None,
    ) -> str:
        """Retrieves a shard iterator. Handles backoff and retry in
        the event of a ProvisionedThroughputExceededException.
        """
        self.logger.debug(
            f'Retrieving shard iterator for shard "{shard_id}" with type "{shard_iterator_type}"'
        )
        try:
            assert int(retry_count or 0) <= self.max_retry_count, "Retries have been exhausted"
            if wait_seconds:
                self.logger.info(f"Waiting {wait_seconds} second(s)...")
                time.sleep(round(wait_seconds, 2))

            get_shard_iterator_input = {
                "StreamName": stream_name,
                "ShardId": shard_id,
                "ShardIteratorType": shard_iterator_type,
            }
            if starting_sequence_number:
                get_shard_iterator_input["StartingSequenceNumber"] = starting_sequence_number
            elif timestamp:
                get_shard_iterator_input["Timestamp"] = timestamp

            return self.client.get_shard_iterator(**get_shard_iterator_input).get("ShardIterator")
        except ClientError as e:
            if e.response["Error"]["Code"] == PROVISIONED_THROUGHPUT_EXCEPTION:
                self.logger.warning(
                    "Encountered a ProvisionedThroughputExceededException. Retrying."
                )
                return self.__get_shard_iterator(
                    stream_name=stream_name,
                    shard_id=shard_id,
                    shard_iterator_type=shard_iterator_type,
                    starting_sequence_number=starting_sequence_number,
                    timestamp=timestamp,
                    wait_seconds=self._next_wait_seconds(wait_seconds),
                    retry_count=self._next_retry_count(retry_count),
                )
            self.logger.error(
                f'Failed to get shard iterator for shard "{shard_id}" with exception: {e}'
            )
            raise e
        except Exception as e:
            self.logger.error(
                f'Failed to get shard iterator for shard "{shard_id}" with exception: {e}'
            )
            raise e

    def __get_records_with_shard_iterator(
        self,
        shard_iterator: Optional[str] = None,
        wait_seconds: Optional[float] = None,
        retry_count: Optional[int] = None,
    ) -> dict:
        """Retrieve records from a Kinesis stream using a shard iterator.
        This function handles backoff and retry in the event of a
        ProvisionedThroughputExceededException.
        """
        try:
            assert int(retry_count or 0) <= self.max_retry_count, "Retries have been exhausted"
            if wait_seconds:
                self.logger.info(f"Waiting {wait_seconds} second(s)...")
                time.sleep(round(wait_seconds, 2))

            return self.client.get_records(ShardIterator=shard_iterator, Limit=10000)
        except ClientError as e:
            if e.response["Error"]["Code"] == PROVISIONED_THROUGHPUT_EXCEPTION:
                self.logger.warning(
                    "Encountered a ProvisionedThroughputExceededException. Retrying."
                )
                return self.__get_records_with_shard_iterator(
                    shard_iterator=shard_iterator,
                    wait_seconds=self._next_wait_seconds(wait_seconds),
                    retry_count=self._next_retry_count(retry_count),
                )
            self.logger.error(f"Failed to get records with exception: {e}")
            raise e
        except Exception as e:
            self.logger.error(f"Failed to get records with exception: {e}")
            raise e

    def _next_wait_seconds(self, current_wait_seconds: Optional[float] = None) -> float:
        if not current_wait_seconds:
            return 1.0
        return current_wait_seconds * self.retry_backoff_rate

    def _next_retry_count(self, current_retry_count: Optional[int] = None) -> int:
        if not current_retry_count:
            return 1
        return current_retry_count + 1



import pytest
import pytz

from botocore.exceptions import ClientError
from datetime import datetime
from psi_kinesis.kinesis_reader import KinesisReader, PROVISIONED_THROUGHPUT_EXCEPTION
from psi_utils.unittest_mock import MockApi
from typing import Tuple


class TestKinesisReader_UnitTests:

    @pytest.mark.unit
    def test_assume_role(self):
        # Set expected values
        expected_role = "RoleName"
        expected_account = "01234567890"
        expected_role_arn = "arn:aws:iam::01234567890:role/RoleName"
        expected_role_session_name = "RoleSession"
        expected_err = Exception("test-err")

        # Set up mocks
        mock_sts_client = MockApi()
        mock_session, _ = self.get_mocks()
        mock_session.on("client", "sts").return_val(mock_sts_client)
        mock_sts_client.on(
            "assume_role", RoleArn=expected_role_arn, RoleSessionName=expected_role_session_name
        ).raise_exception(expected_err)

        # Test assume_role() Exception
        kr = KinesisReader(aws_session=mock_session)
        with pytest.raises(Exception) as e_info:
            kr.assume_role(
                expected_role, expected_account, role_session_name=expected_role_session_name
            )
        assert e_info.value == expected_err
        mock_sts_client.assert_expectations()
        mock_session.assert_expectations()

    @pytest.mark.unit
    def test_does_stream_exist(self):
        # Set expected values
        expected_stream_name = "kinesis-stream-name"
        expected_resp = {"StreamDescription": {"StreamName": expected_stream_name}}

        # Set up mocks
        mock_session, mock_kinesis_client = self.get_mocks()
        mock_kinesis_client.on("describe_stream", StreamName=expected_stream_name).return_val(
            expected_resp
        )

        # Test does_stream_exist()
        kr = KinesisReader(aws_session=mock_session)
        assert kr.does_stream_exist(expected_stream_name)
        mock_kinesis_client.assert_expectations()
        mock_session.assert_expectations()

    @pytest.mark.unit
    def test_does_stream_exist_resource_not_found(self):
        # Set expected values
        expected_stream_name = "kinesis-stream-name"
        expected_err = ClientError({"Error": {"Code": "ResourceNotFoundException"}}, "test")

        # Set up mocks
        mock_session, mock_kinesis_client = self.get_mocks()
        mock_kinesis_client.on("describe_stream", StreamName=expected_stream_name).raise_exception(
            expected_err
        )

        # Test does_stream_exist() ResourceNotFoundException
        kr = KinesisReader(aws_session=mock_session)
        assert not kr.does_stream_exist(expected_stream_name)
        mock_kinesis_client.assert_expectations()
        mock_session.assert_expectations()

    @pytest.mark.unit
    def test_does_stream_exist_client_error(self):
        # Set expected values
        expected_stream_name = "kinesis-stream-name"
        expected_err = ClientError({"Error": {"Code": "ClientError"}}, "test")

        # Set up mocks
        mock_session, mock_kinesis_client = self.get_mocks()
        mock_kinesis_client.on("describe_stream", StreamName=expected_stream_name).raise_exception(
            expected_err
        )

        # Test does_stream_exist() ClientError
        kr = KinesisReader(aws_session=mock_session)
        with pytest.raises(ClientError) as e_info:
            kr.does_stream_exist(expected_stream_name)
        assert e_info.value == expected_err
        mock_kinesis_client.assert_expectations()
        mock_session.assert_expectations()

    @pytest.mark.unit
    def test_does_stream_exist_error(self):
        # Set expected values
        expected_stream_name = "kinesis-stream-name"
        expected_err = Exception("test-exception")

        # Set up mocks
        mock_session, mock_kinesis_client = self.get_mocks()
        mock_kinesis_client.on("describe_stream", StreamName=expected_stream_name).raise_exception(
            expected_err
        )

        # Test does_stream_exist() Exception
        kr = KinesisReader(aws_session=mock_session)
        with pytest.raises(Exception) as e_info:
            kr.does_stream_exist(expected_stream_name)
        assert e_info.value == expected_err
        mock_kinesis_client.assert_expectations()
        mock_session.assert_expectations()

    @pytest.mark.unit
    def test_list_shards(self):
        # Set expected values
        expected_stream_name = "kinesis-stream-name"
        expected_shard_filter = {
            "Type": "FROM_TIMESTAMP",
            "Timestamp": datetime(2022, 4, 1, tzinfo=pytz.UTC),
        }
        expected_shards = [{"ShardId": "shard-id-000001"}]

        # Set up mocks
        mock_session, mock_kinesis_client = self.get_mocks()
        mock_kinesis_client.on(
            "list_shards", StreamName=expected_stream_name, ShardFilter=expected_shard_filter
        ).return_val({"Shards": expected_shards})

        # Test list_shards()
        kr = KinesisReader(aws_session=mock_session)
        shards = kr.list_shards(expected_stream_name, shard_filter=expected_shard_filter)
        assert shards == expected_shards
        mock_kinesis_client.assert_expectations()
        mock_session.assert_expectations()

    @pytest.mark.unit
    def test_list_shards_error(self):
        # Set expected values
        expected_stream_name = "kinesis-stream-name"
        expected_err = Exception("test-exception")

        # Set up mocks
        mock_session, mock_kinesis_client = self.get_mocks()
        mock_kinesis_client.on("list_shards", StreamName=expected_stream_name).raise_exception(
            expected_err
        )

        # Test list_shards() Exception
        kr = KinesisReader(aws_session=mock_session)
        with pytest.raises(Exception) as e_info:
            kr.list_shards(expected_stream_name)
        assert e_info.value == expected_err
        mock_kinesis_client.assert_expectations()
        mock_session.assert_expectations()

    @pytest.mark.unit
    def test_get_records(self):
        # Set expected values
        expected_stream_name = "kinesis-stream-name"
        expected_shard_id = "shard-id-000001"
        expected_shard_iterator = "shard-iterator-id"
        expected_records = [{"a": 1}, {"b": 2}, {"c": 3}]

        # Set up mocks
        mock_session, mock_kinesis_client = self.get_mocks()
        mock_kinesis_client.on(
            "get_shard_iterator",
            StreamName=expected_stream_name,
            ShardId=expected_shard_id,
            ShardIteratorType="TRIM_HORIZON",
        ).return_val({"ShardIterator": expected_shard_iterator})
        mock_kinesis_client.on(
            "get_records", ShardIterator=expected_shard_iterator, Limit=10000
        ).return_val({"Records": expected_records, "MillisBehindLatest": 0})

        # Test get_records() with a sequence number but no shard iterator type
        kr = KinesisReader(aws_session=mock_session)
        with pytest.raises(AssertionError) as e_info:
            kr.get_records(
                expected_stream_name,
                shard_id=expected_shard_id,
                starting_sequence_number="1234567890123456789",
            )
        assert (
            str(e_info.value)
            == "Cannot specify a starting sequence number or timestamp if shard iteratory type is one of [TRIM_HORIZON,LATEST]"
        )

        # Test get_records() with a timestamp but no shard iterator type
        kr = KinesisReader(aws_session=mock_session)
        with pytest.raises(AssertionError) as e_info:
            kr.get_records(
                expected_stream_name,
                shard_id=expected_shard_id,
                timestamp=datetime(2022, 4, 1, tzinfo=pytz.UTC),
            )
        assert (
            str(e_info.value)
            == "Cannot specify a starting sequence number or timestamp if shard iteratory type is one of [TRIM_HORIZON,LATEST]"
        )

        # Test get_records()
        kr = KinesisReader(aws_session=mock_session)
        records = kr.get_records(expected_stream_name, shard_id=expected_shard_id)
        assert records == {expected_shard_id: expected_records}
        mock_kinesis_client.assert_expectations()
        mock_session.assert_expectations()

    @pytest.mark.unit
    def test_get_records_at_sequence_number(self):
        # Set expected values
        expected_stream_name = "kinesis-stream-name"
        expected_shard_id = "shard-id-000001"
        expected_shard_iterator = "shard-iterator-id"
        expected_shard_iterator_type = "AT_SEQUENCE_NUMBER"
        expected_sequence_number = "1234567890123456789"
        expected_records = [{"a": 1}, {"b": 2}, {"c": 3}]

        # Set up mocks
        mock_session, mock_kinesis_client = self.get_mocks()
        mock_kinesis_client.on(
            "get_shard_iterator",
            StreamName=expected_stream_name,
            ShardId=expected_shard_id,
            ShardIteratorType=expected_shard_iterator_type,
            StartingSequenceNumber=expected_sequence_number,
        ).return_val({"ShardIterator": expected_shard_iterator})
        mock_kinesis_client.on(
            "get_records", ShardIterator=expected_shard_iterator, Limit=10000
        ).return_val({"Records": expected_records, "MillisBehindLatest": 0})

        # Test get_records() AT_SEQUENCE_NUMBER
        kr = KinesisReader(aws_session=mock_session)
        records = kr.get_records(
            expected_stream_name,
            shard_id=expected_shard_id,
            shard_iterator_type=expected_shard_iterator_type,
            starting_sequence_number=expected_sequence_number,
        )
        assert records == {expected_shard_id: expected_records}
        mock_kinesis_client.assert_expectations()
        mock_session.assert_expectations()

    @pytest.mark.unit
    def test_get_records_at_timestamp(self):
        # Set expected values
        expected_stream_name = "kinesis-stream-name"
        expected_shard_id = "shard-id-000001"
        expected_shard_iterator = "shard-iterator-id"
        expected_shard_iterator_type = "AT_TIMESTAMP"
        expected_timestamp = datetime(2022, 4, 1, tzinfo=pytz.UTC)
        expected_records = [{"a": 1}, {"b": 2}, {"c": 3}]

        # Set up mocks
        mock_session, mock_kinesis_client = self.get_mocks()
        mock_kinesis_client.on(
            "get_shard_iterator",
            StreamName=expected_stream_name,
            ShardId=expected_shard_id,
            ShardIteratorType=expected_shard_iterator_type,
            Timestamp=expected_timestamp,
        ).return_val({"ShardIterator": expected_shard_iterator})
        mock_kinesis_client.on(
            "get_records", ShardIterator=expected_shard_iterator, Limit=10000
        ).return_val({"Records": expected_records, "MillisBehindLatest": 0})

        # Test get_records() AT_TIMESTAMP
        kr = KinesisReader(aws_session=mock_session)
        records = kr.get_records(
            expected_stream_name,
            shard_id=expected_shard_id,
            shard_iterator_type=expected_shard_iterator_type,
            timestamp=expected_timestamp,
        )
        assert records == {expected_shard_id: expected_records}
        mock_kinesis_client.assert_expectations()
        mock_session.assert_expectations()

    @pytest.mark.unit
    def test_get_records_for_entire_stream(self):
        # Set expected values
        expected_stream_name = "kinesis-stream-name"
        expected_shard_iterator = "shard-iterator-id"
        expected_records = [{"a": 1}, {"b": 2}, {"c": 3}]
        expected_shard_id = "shard-id-000001"

        # Set up mocks
        mock_session, mock_kinesis_client = self.get_mocks()
        mock_kinesis_client.on("list_shards", StreamName=expected_stream_name).return_val(
            {"Shards": [{"ShardId": expected_shard_id}]}
        )
        mock_kinesis_client.on(
            "get_shard_iterator",
            StreamName=expected_stream_name,
            ShardId=expected_shard_id,
            ShardIteratorType="TRIM_HORIZON",
        ).return_val({"ShardIterator": expected_shard_iterator})
        mock_kinesis_client.on(
            "get_records", ShardIterator=expected_shard_iterator, Limit=10000
        ).return_val({"Records": expected_records, "MillisBehindLatest": 0})

        # Test get_records() for an entire Kinesis stream
        kr = KinesisReader(aws_session=mock_session)
        records = kr.get_records(expected_stream_name)
        assert records == {expected_shard_id: expected_records}
        mock_kinesis_client.assert_expectations()
        mock_session.assert_expectations()

    @pytest.mark.unit
    def test_get_records_with_shard_iterator(self):
        # Set expected values
        fst_shard_iterator = "first-shard-iterator-id"
        snd_shard_iterator = "second-shard-iterator-id"
        fst_records = [{"a": 1}, {"b": 2}]
        snd_records = [{"c": 3}, {"d": 4}]
        fst_resp = {
            "Records": fst_records,
            "NextShardIterator": snd_shard_iterator,
            "MillisBehindLatest": 123,
        }
        snd_resp = {
            "Records": snd_records,
            "NextShardIterator": "third-shard-iterator-id",
            "MillisBehindLatest": 0,
        }
        expected_records = fst_records + snd_records

        # Set up mocks
        mock_session, mock_kinesis_client = self.get_mocks()
        mock_kinesis_client.on(
            "get_records", ShardIterator=fst_shard_iterator, Limit=10000
        ).return_val(fst_resp)
        mock_kinesis_client.on(
            "get_records", ShardIterator=snd_shard_iterator, Limit=10000
        ).return_val(snd_resp)

        # Test get_records_with_shard_iterator()
        kr = KinesisReader(aws_session=mock_session)
        records = kr.get_records_with_shard_iterator(fst_shard_iterator)
        assert records == expected_records
        mock_kinesis_client.assert_expectations()
        mock_session.assert_expectations()

    @pytest.mark.unit
    def test__get_shard_iterator(self):
        # Set expected values
        expected_stream_name = "kinesis-stream-name"
        expected_shard_id = "shard-id-000001"
        expected_shard_iterator_type = "AT_SEQUENCE_NUMBER"
        expected_sequence_number = "1234567890123456789"
        expected_shard_iterator = "shard-iterator-id"
        expected_resp = {"ShardIterator": expected_shard_iterator}

        # Set up mocks
        mock_session, mock_kinesis_client = self.get_mocks()
        mock_kinesis_client.on(
            "get_shard_iterator",
            StreamName=expected_stream_name,
            ShardId=expected_shard_id,
            ShardIteratorType=expected_shard_iterator_type,
            StartingSequenceNumber=expected_sequence_number,
        ).return_val(expected_resp)

        # Test __get_shard_iterator()
        kr = KinesisReader(aws_session=mock_session)
        shard_iterator = kr._KinesisReader__get_shard_iterator(
            expected_stream_name,
            expected_shard_id,
            expected_shard_iterator_type,
            starting_sequence_number=expected_sequence_number,
        )
        assert shard_iterator == expected_shard_iterator
        mock_kinesis_client.assert_expectations()
        mock_session.assert_expectations()

    @pytest.mark.unit
    def test__get_shard_iterator_throughput_exception(self):
        # Set expected values
        expected_stream_name = "kinesis-stream-name"
        expected_shard_id = "shard-id-000001"
        expected_shard_iterator_type = "AT_TIMESTAMP"
        expected_timestamp = datetime(2022, 4, 1, tzinfo=pytz.UTC)
        expected_err = ClientError({"Error": {"Code": PROVISIONED_THROUGHPUT_EXCEPTION}}, "test")

        # Set up mocks
        mock_session, mock_kinesis_client = self.get_mocks()
        mock_kinesis_client.on(
            "get_shard_iterator",
            StreamName=expected_stream_name,
            ShardId=expected_shard_id,
            ShardIteratorType=expected_shard_iterator_type,
            Timestamp=expected_timestamp,
        ).raise_exception(expected_err)

        # Test __get_shard_iterator() ProvisionedThroughputExceededException
        kr = KinesisReader(aws_session=mock_session, max_retry_count=1)

        with pytest.raises(AssertionError) as e_info:
            kr._KinesisReader__get_shard_iterator(
                expected_stream_name,
                expected_shard_id,
                expected_shard_iterator_type,
                timestamp=expected_timestamp,
            )
        assert str(e_info.value) == "Retries have been exhausted"
        mock_kinesis_client.assert_expectations()
        mock_session.assert_expectations()

    @pytest.mark.unit
    def test__get_shard_iterator_client_error(self):
        # Set expected values
        expected_stream_name = "kinesis-stream-name"
        expected_shard_id = "shard-id-000001"
        expected_shard_iterator_type = "AT_TIMESTAMP"
        expected_timestamp = datetime(2022, 4, 1, tzinfo=pytz.UTC)
        expected_err = ClientError({"Error": {"Code": "ClientError"}}, "test")

        # Set up mocks
        mock_session, mock_kinesis_client = self.get_mocks()
        mock_kinesis_client.on(
            "get_shard_iterator",
            StreamName=expected_stream_name,
            ShardId=expected_shard_id,
            ShardIteratorType=expected_shard_iterator_type,
            Timestamp=expected_timestamp,
        ).raise_exception(expected_err)

        # Test __get_shard_iterator() ClientError
        kr = KinesisReader(aws_session=mock_session, max_retry_count=1)

        with pytest.raises(ClientError) as e_info:
            kr._KinesisReader__get_shard_iterator(
                expected_stream_name,
                expected_shard_id,
                expected_shard_iterator_type,
                timestamp=expected_timestamp,
            )
        assert e_info.value == expected_err
        mock_kinesis_client.assert_expectations()
        mock_session.assert_expectations()

    @pytest.mark.unit
    def test__get_shard_iterator_error(self):
        # Set expected values
        expected_stream_name = "kinesis-stream-name"
        expected_shard_id = "shard-id-000001"
        expected_shard_iterator_type = "AT_TIMESTAMP"
        expected_timestamp = datetime(2022, 4, 1, tzinfo=pytz.UTC)
        expected_err = Exception("test-exception")

        # Set up mocks
        mock_session, mock_kinesis_client = self.get_mocks()
        mock_kinesis_client.on(
            "get_shard_iterator",
            StreamName=expected_stream_name,
            ShardId=expected_shard_id,
            ShardIteratorType=expected_shard_iterator_type,
            Timestamp=expected_timestamp,
        ).raise_exception(expected_err)

        # Test __get_shard_iterator() Exception
        kr = KinesisReader(aws_session=mock_session, max_retry_count=1)

        with pytest.raises(Exception) as e_info:
            kr._KinesisReader__get_shard_iterator(
                expected_stream_name,
                expected_shard_id,
                expected_shard_iterator_type,
                timestamp=expected_timestamp,
            )
        assert e_info.value == expected_err
        mock_kinesis_client.assert_expectations()
        mock_session.assert_expectations()

    @pytest.mark.unit
    def test__get_records_with_shard_iterator(self):
        # Set expected values
        expected_shard_iterator = "shard-iterator-id"
        expected_resp = {"Records": [], "NextShardIterator": None, "MillisBehindLatest": 123}

        # Set up mocks
        mock_session, mock_kinesis_client = self.get_mocks()
        mock_kinesis_client.on(
            "get_records", ShardIterator=expected_shard_iterator, Limit=10000
        ).return_val(expected_resp)

        # Test __get_records_with_shard_iterator()
        kr = KinesisReader(aws_session=mock_session)
        resp = kr._KinesisReader__get_records_with_shard_iterator(expected_shard_iterator)
        assert resp == expected_resp
        mock_kinesis_client.assert_expectations()
        mock_session.assert_expectations()

    @pytest.mark.unit
    def test__get_records_with_shard_iterator_throughput_exception(self):
        # Set expected values
        expected_shard_iterator = "shard-iterator-id"
        expected_err = ClientError({"Error": {"Code": PROVISIONED_THROUGHPUT_EXCEPTION}}, "test")

        # Set up mocks
        mock_session, mock_kinesis_client = self.get_mocks()
        mock_kinesis_client.on(
            "get_records", ShardIterator=expected_shard_iterator, Limit=10000
        ).raise_exception(expected_err)

        # Test __get_records_with_shard_iterator() ProvisionedThroughputExceededException
        kr = KinesisReader(aws_session=mock_session, max_retry_count=1)

        with pytest.raises(AssertionError) as e_info:
            kr._KinesisReader__get_records_with_shard_iterator(expected_shard_iterator)
        assert str(e_info.value) == "Retries have been exhausted"
        mock_kinesis_client.assert_expectations()
        mock_session.assert_expectations()

    @pytest.mark.unit
    def test__get_records_with_shard_iterator_client_error(self):
        # Set expected values
        expected_shard_iterator = "shard-iterator-id"
        expected_err = ClientError({"Error": {"Code": "ClientError"}}, "test")

        # Set up mocks
        mock_session, mock_kinesis_client = self.get_mocks()
        mock_kinesis_client.on(
            "get_records", ShardIterator=expected_shard_iterator, Limit=10000
        ).raise_exception(expected_err)

        # Test __get_records_with_shard_iterator() ClientError
        kr = KinesisReader(aws_session=mock_session, max_retry_count=1)

        with pytest.raises(ClientError) as e_info:
            kr._KinesisReader__get_records_with_shard_iterator(expected_shard_iterator)
        assert e_info.value == expected_err
        mock_kinesis_client.assert_expectations()
        mock_session.assert_expectations()

    @pytest.mark.unit
    def test__get_records_with_shard_iterator_error(self):
        # Set expected values
        expected_shard_iterator = "shard-iterator-id"
        expected_err = Exception("test exception")

        # Set up mocks
        mock_session, mock_kinesis_client = self.get_mocks()
        mock_kinesis_client.on(
            "get_records", ShardIterator=expected_shard_iterator, Limit=10000
        ).raise_exception(expected_err)

        # Test __get_records_with_shard_iterator() Exception
        kr = KinesisReader(aws_session=mock_session, max_retry_count=1)

        with pytest.raises(Exception) as e_info:
            kr._KinesisReader__get_records_with_shard_iterator(expected_shard_iterator)
        assert e_info.value == expected_err
        mock_kinesis_client.assert_expectations()
        mock_session.assert_expectations()

    @pytest.mark.unit
    def test_next_wait_seconds(self):
        retry_backoff_rate = 2.5
        mock_session, _ = self.get_mocks()
        kr = KinesisReader(aws_session=mock_session, retry_backoff_rate=retry_backoff_rate)
        assert kr._next_wait_seconds(None) == 1.0
        assert kr._next_wait_seconds(1.0) == retry_backoff_rate
        assert kr._next_wait_seconds(retry_backoff_rate) == 6.25
        assert kr._next_wait_seconds(10) == 25

    @pytest.mark.unit
    def test_next_retry_count(self):
        mock_session, _ = self.get_mocks()
        kr = KinesisReader(aws_session=mock_session)
        assert kr._next_retry_count(None) == 1
        assert kr._next_retry_count(10) == 11

    def get_mocks(self) -> Tuple[MockApi, MockApi]:
        mock_kinesis_client = MockApi()
        mock_session = MockApi()
        mock_session.on("client", "kinesis").return_val(mock_kinesis_client)
        return mock_session, mock_kinesis_client



import pytest
import pytz
import uuid

from botocore.exceptions import ClientError
from datetime import datetime
from psi_kinesis.kinesis_reader import PROVISIONED_THROUGHPUT_EXCEPTION
from psi_kinesis.kinesis_writer import KinesisWriter, MAX_INVALID_RECORDS, MAX_PUT_RECORDS_BATCH
from psi_utils.unittest_mock import MockApi
from typing import Tuple


class TestKinesisWriter_UnitTests:

    @pytest.mark.unit
    def test_put_records(self):
        # Set expected values
        expected_records = [{"Data": b"Some data", "PartitionKey": "partitionkey"}]
        expected_stream_name = "kinesis-stream-name"
        expected_sequence_number = "1234567890123456789"
        expected_shard_id = "shard-000001"
        expected_resp = [{"SequenceNumber": expected_sequence_number, "ShardId": expected_shard_id}]
        expected_put_records_batch_resp = {"Records": expected_resp}

        # Set up mocks
        mock_session, mock_kinesis_client = self.get_mocks()
        mock_kinesis_client.on(
            "put_records", Records=expected_records, StreamName=expected_stream_name
        ).return_val(expected_put_records_batch_resp)

        # Test put_records()
        kw = KinesisWriter(aws_session=mock_session)
        resp = kw.put_records(stream_name=expected_stream_name, records=expected_records)
        assert resp == expected_resp
        mock_kinesis_client.assert_expectations()
        mock_session.assert_expectations()

    @pytest.mark.unit
    def test_put_records_quit_on_err(self):
        # Set expected values
        expected_records = [{"Data": b"Some data", "PartitionKey": "partitionkey"}]
        expected_stream_name = "kinesis-stream-name"
        expected_put_records_batch_resp = {
            "FailedRecordCount": 1,
            "Records": [{"ErrorMessage": "Some error message"}],
        }

        # Set up mocks
        mock_session, mock_kinesis_client = self.get_mocks()
        mock_kinesis_client.on(
            "put_records", Records=expected_records, StreamName=expected_stream_name
        ).return_val(expected_put_records_batch_resp)

        # Test put_records() failure with quit_on_error=True
        kw = KinesisWriter(aws_session=mock_session)
        with pytest.raises(AssertionError) as e_info:
            kw.put_records(
                stream_name=expected_stream_name, records=expected_records, quit_on_error=True
            )
        assert str(e_info.value).startswith("1 record(s) failed to write")
        mock_kinesis_client.assert_expectations()
        mock_session.assert_expectations()

    @pytest.mark.unit
    def test_put_records_no_quit_on_err(self):
        # Set expected values
        expected_records = [
            {"Data": b"Some data", "PartitionKey": f"partitionkey{i}"}
            for i in range(MAX_PUT_RECORDS_BATCH * 3)
        ]
        expected_stream_name = "kinesis-stream-name"
        fst_expected_err = Exception("first-test-exception")
        snd_expected_err = Exception("second-test-exception")
        thd_expected_err = Exception("third-test-exception")
        expected_exception = Exception([fst_expected_err, snd_expected_err, thd_expected_err])

        # Set up mocks
        mock_session, mock_kinesis_client = self.get_mocks()
        mock_kinesis_client.on(
            "put_records",
            Records=expected_records[:MAX_PUT_RECORDS_BATCH],
            StreamName=expected_stream_name,
        ).raise_exception(fst_expected_err)
        mock_kinesis_client.on(
            "put_records",
            Records=expected_records[MAX_PUT_RECORDS_BATCH:-MAX_PUT_RECORDS_BATCH],
            StreamName=expected_stream_name,
        ).raise_exception(snd_expected_err)
        mock_kinesis_client.on(
            "put_records",
            Records=expected_records[-MAX_PUT_RECORDS_BATCH:],
            StreamName=expected_stream_name,
        ).raise_exception(thd_expected_err)

        # Test put_records() failure Exceptions quit_on_error=False
        kw = KinesisWriter(aws_session=mock_session)
        with pytest.raises(Exception) as e_info:
            kw.put_records(
                stream_name=expected_stream_name, records=expected_records, quit_on_error=False
            )
        assert str(e_info.value) == str(expected_exception)
        mock_kinesis_client.assert_expectations()
        mock_session.assert_expectations()

    @pytest.mark.unit
    def test_put_records_error(self):
        # Set expected values
        expected_records = [{"Data": b"Some data", "PartitionKey": "partitionkey"}]
        expected_stream_name = "kinesis-stream-name"
        expected_err = Exception("test-exception")

        # Set up mocks
        mock_session, mock_kinesis_client = self.get_mocks()
        mock_kinesis_client.on(
            "put_records", Records=expected_records, StreamName=expected_stream_name
        ).raise_exception(expected_err)

        # Test put_records() Exception with quit_on_error=True
        kw = KinesisWriter(aws_session=mock_session)
        with pytest.raises(Exception) as e_info:
            kw.put_records(
                stream_name=expected_stream_name, records=expected_records, quit_on_error=True
            )
        assert e_info.value == expected_err
        mock_kinesis_client.assert_expectations()
        mock_session.assert_expectations()

    @pytest.mark.unit
    def test__put_records_batch(self):
        # Set expected values
        expected_batch = [{"a": 1}, {"b": 2}, {"c": 3}]
        expected_stream_name = "kinesis-stream-name"
        expected_resp = {"Records": []}

        # Set up mocks
        mock_session, mock_kinesis_client = self.get_mocks()
        mock_kinesis_client.on(
            "put_records", Records=expected_batch, StreamName=expected_stream_name
        ).return_val(expected_resp)

        # Test __put_records_batch()
        kw = KinesisWriter(aws_session=mock_session)
        resp = kw._KinesisWriter__put_records_batch(
            batch=expected_batch, stream_name=expected_stream_name
        )
        assert resp == expected_resp
        mock_kinesis_client.assert_expectations()
        mock_session.assert_expectations()

    @pytest.mark.unit
    def test__put_records_batch_throughput_exception(self):
        # Set expected values
        expected_batch = [{"a": 1}, {"b": 2}, {"c": 3}]
        expected_stream_name = "kinesis-stream-name"
        expected_resp = {
            "FailedRecordCount": 1,
            "Records": [{"ErrorCode": PROVISIONED_THROUGHPUT_EXCEPTION}],
        }

        # Set up mocks
        mock_session, mock_kinesis_client = self.get_mocks()
        mock_kinesis_client.on(
            "put_records", Records=expected_batch, StreamName=expected_stream_name
        ).return_val(expected_resp)

        # Test __put_records_batch() ProvisionedThroughputExceededException
        kw = KinesisWriter(aws_session=mock_session, max_retry_count=1)
        with pytest.raises(AssertionError) as e_info:
            kw._KinesisWriter__put_records_batch(
                batch=expected_batch, stream_name=expected_stream_name
            )
        assert str(e_info.value) == "Retries have been exhausted"
        mock_kinesis_client.assert_expectations()
        mock_session.assert_expectations()

    @pytest.mark.unit
    def test__put_records_batch_failure(self):
        # Set expected values
        expected_batch = [{"a": 1}, {"b": 2}, {"c": 3}]
        expected_stream_name = "kinesis-stream-name"
        expected_resp = {"FailedRecordCount": 1, "Records": [{"ErrorCode": "SomeOtherFailure"}]}

        # Set up mocks
        mock_session, mock_kinesis_client = self.get_mocks()
        mock_kinesis_client.on(
            "put_records", Records=expected_batch, StreamName=expected_stream_name
        ).return_val(expected_resp)

        # Test __put_records_batch() failures
        kw = KinesisWriter(aws_session=mock_session, max_retry_count=1)
        resp = kw._KinesisWriter__put_records_batch(
            batch=expected_batch, stream_name=expected_stream_name
        )
        assert resp == expected_resp
        mock_kinesis_client.assert_expectations()
        mock_session.assert_expectations()

    @pytest.mark.unit
    def test__put_records_batch_error(self):
        # Set expected values
        expected_batch = [{"a": 1}, {"b": 2}, {"c": 3}]
        expected_stream_name = "kinesis-stream-name"
        expected_err = Exception("test-exception")

        # Set up mocks
        mock_session, mock_kinesis_client = self.get_mocks()
        mock_kinesis_client.on(
            "put_records", Records=expected_batch, StreamName=expected_stream_name
        ).raise_exception(expected_err)

        # Test __put_records_batch() Exception
        kw = KinesisWriter(aws_session=mock_session, max_retry_count=1)
        with pytest.raises(Exception) as e_info:
            kw._KinesisWriter__put_records_batch(
                batch=expected_batch, stream_name=expected_stream_name
            )
        assert e_info.value == expected_err
        mock_kinesis_client.assert_expectations()
        mock_session.assert_expectations()

    @pytest.mark.unit
    def test__partition_put_records(self):
        mock_session, _ = self.get_mocks()
        kw = KinesisWriter(aws_session=mock_session)

        records_with_partition = kw._KinesisWriter__partition_put_records(
            records=[{"Data": b"Some data"}]
        )
        assert len(records_with_partition) == 1
        assert "PartitionKey" in records_with_partition[0]
        assert isinstance(records_with_partition[0]["PartitionKey"], str)
        mock_session.assert_expectations()

    @pytest.mark.unit
    def test__batch_put_records(self):
        mock_session, _ = self.get_mocks()
        kw = KinesisWriter(aws_session=mock_session)

        batched = kw._KinesisWriter__batch_put_records([{"a": 1}, {"b": 1}, {"c": 1}], batch_size=2)
        assert list(batched) == [[{"a": 1}, {"b": 1}], [{"c": 1}]]
        mock_session.assert_expectations()

    @pytest.mark.unit
    def test__validate_put_records(self):
        # Generate records to validate
        invalid_records = [
            {
                "Data": b"Some data",
                "PartitionKey": uuid.uuid4().hex,
                "ExplicitHashKey": uuid.uuid4().hex,
            }
            for _ in range(MAX_INVALID_RECORDS + 2)
        ]
        valid_records = [{"Data": b"Some data"} for _ in range(5)]
        records = valid_records + invalid_records

        # Set expected values
        expected_invalid_records = [
            {
                "record": r,
                "invalid_reason": "Expected either record.PartitionKey or record.ExplicitHashKey, not both",
            }
            for r in invalid_records[:-2]
        ]

        # Test __validate_put_records()
        mock_session, _ = self.get_mocks()
        kw = KinesisWriter(aws_session=mock_session)
        actual_invalid_records = kw._KinesisWriter__validate_put_records(records)
        assert actual_invalid_records == expected_invalid_records
        mock_session.assert_expectations()

    @pytest.mark.unit
    def test__validate_put_record(self):
        mock_session, _ = self.get_mocks()
        kw = KinesisWriter(aws_session=mock_session)

        # Test has PartitionKey
        valid, invalid_reason = kw._KinesisWriter__validate_put_record(
            {"Data": b"Some data", "PartitionKey": "partitionkey"}
        )
        assert valid
        assert not invalid_reason

        # Test has ExplicitHashKey
        valid, invalid_reason = kw._KinesisWriter__validate_put_record(
            {"Data": b"Some data", "ExplicitHashKey": "explicithashkey"}
        )
        assert valid
        assert not invalid_reason

        # Test invalid record
        valid, invalid_reason = kw._KinesisWriter__validate_put_record(
            {
                "Data": b"Some data",
                "PartitionKey": "partitionkey",
                "ExplicitHashKey": "explicithashkey",
            }
        )
        assert not valid
        assert (
            invalid_reason
            == "Expected either record.PartitionKey or record.ExplicitHashKey, not both"
        )

    @pytest.mark.unit
    def test__put_records_resp_has_throughput_exception(self):
        mock_session, _ = self.get_mocks()
        kw = KinesisWriter(aws_session=mock_session)
        resp = kw._KinesisWriter__put_records_resp_has_throughput_exception(
            {
                "FailedRecordCount": 2,
                "Records": [
                    {"ErrorCode": PROVISIONED_THROUGHPUT_EXCEPTION},
                    {"ErrorCode": "SomeOtherError"},
                ],
            }
        )
        has_throughput_exception, has_other_exception = resp[0], resp[1]
        assert has_throughput_exception
        assert has_other_exception
        mock_session.assert_expectations()

    def get_mocks(self) -> Tuple[MockApi, MockApi]:
        mock_kinesis_client = MockApi()
        mock_session = MockApi()
        mock_session.on("client", "kinesis").return_val(mock_kinesis_client)
        return mock_session, mock_kinesis_client
